#ifndef TFIO_RIFF_H
#define TFIO_RIFF_H

#include <cstring>
#include <iostream>
#include <fstream>
#include "data.h"
#include "common.h"

namespace harkio {

/**********************************************************************/
//! RIFF抽象基底クラス
/*!

*/
/**********************************************************************/
class RIFF{
protected:
	//! RIFFへDataを追加する前に、エラーチェックを行う
	virtual void addDataCheck(Data *data) = 0;

	//! 派生クラスの識別子を取得する
	virtual const char *getIdentifierFromBase() = 0;

	//! 書き込む前にデータが揃っているか確認する
	/*! サイズ以外のデータの整合性をチェックする(包含データを再帰風に呼び出している)
	 */
	virtual bool isComplete()= 0;

	//! サイズフィールド
	int size;
	//! 保持しているDataの配列
	std::vector<Data*> data;
public:
	RIFF(){
	}
	virtual ~RIFF(){
		for(unsigned i=0; i < data.size(); i++) {
			delete data[i];
		}
	}

	//! ファイルからこのクラスのデータを読みとる
	virtual void readFile(std::string fullpath_file) = 0;

	//! ファイルへデータを書き込む
	virtual void writeFile(std::string fullpath_file) = 0;

	//! サイズフィールドを書き込む関数
	/*!
		包含データを再帰風に呼び出している
		(isComplete()にてデータ構造が正しい事を確認してから実行する事)
	 */
	void setSize();

	//! サイズフィールドの値を取得する関数
	int getSize();

	//! サイズヘッダーのサイズを含む自身のクラスのサイズを取得する関数。親の階層がsetSizeする際に必要になる
	int getClassSize();

	//! RIFFへDataの追加を行う
	virtual void addData(Data *data);

	//! RIFFが保持しているDataを取得する
	/*! \param identifier 取得するDataの識別子
	 *  識別子が未指定の場合は、最初にaddDataされたDataが戻る
	 */
	virtual Data *getData(std::string identifier="");

	//! RIFFが保持しているDataの全てを配列で取得する
	std::vector<Data*> getDatas();
};

/**********************************************************************/
//! HARKクラス
/*!

 */
/**********************************************************************/
class HARK : public RIFF {
private:
	//! 基底クラスからこのクラスの識別子を取得する
	const char *getIdentifierFromBase() {
		return getIdentifier();
	}
	//! RIFFへDataを追加する前に、エラーチェックを行う
	void addDataCheck(Data *data) {
		if (!dynamic_cast<HGTF*>(data)) {
//			HARKIO_THROW_EXCEPTION("Data type wasn't HGTF");
		}
		if(RIFF::data.size() !=0) {
			HARKIO_THROW_ERROR("Data's object already exists.(Multiple Data is future extension.)");
		}
	}
public:
	//! 書き込む前にデータが揃っているか確認する
	/*! サイズ以外のデータの整合性をチェックする(包含データを再帰風に呼び出している)
	 */
	bool isComplete();

public:
	HARK() {}
	~HARK() {
	}
	//! 引数に与えられたファイルの、RIFF識別子を取得する
	/*! 主に、ファイルを読み込む前に識別子を読み取りたい場合に使用する
	 * \param fullpath_file 判別対象のファイル名
	 * \return 判別結果
	 * \retval HARKIO_ERROR 判別不能(ファイルエラー等)
	 * \retval HARKIO_TRUE  このクラスの識別子と同じ
	 * \retval HARKIO_FALSE このクラスの識別子ではない
	 */
	static HARKIO_RET checkIdentifier(std::string fullpath_file) {
		char buf[strlen(getIdentifier())+1];
		memset(buf, '\0', sizeof(buf));
		// Open
		std::ifstream fs(fullpath_file.c_str(), std::ios::binary);
		if (fs.fail()) {
			HARKIO_SET_ERROR(std::string("Can't open file :") + fullpath_file);
			return HARKIO_ERROR;
		}
		// 識別子
		fs.read((char*)buf, strlen(getIdentifier()));
		if(fs.fail()) {
			HARKIO_SET_ERROR("Can't read HARK's identifier.");
			return HARKIO_ERROR;
		}
		if(strcmp(buf, getIdentifier())) {
			std::stringstream ss;
			ss << "Identifier is different from HARK's one:" << buf;
			HARKIO_SET_ERROR(ss.str());
			return HARKIO_FALSE;
		}
		return HARKIO_TRUE;
	}
	//! このクラスの識別子を取得する
	static const char *getIdentifier() {
		// 識別子
		static const char *identifier=NULL;
		if(!identifier) {
			identifier ="HARK";
		}
		return identifier;
	}
	//! ファイルへデータを書き込む
	void writeFile(std::string fullpath_file);
	//! ファイルからこのクラスのデータを読みとる
	void readFile(std::string fullpath_file);
};

} // namespace

#endif

